==> ubee512-1.0.0.tar.gz: ubee512-1.0.0/src/sound.c <==
//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                               Sound module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// This module is used to emulate the internal speaker using SDL.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.00 - 14 July 2007, S.J.Kay
// - Created a new file to implement the sound emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#ifdef use_audio

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "SDL.h"
#include "sound.h"
#include "ubee512.h"
#include "mz80.h"

#define AUDIO_U8_ON_VOLUME 128+16
#define AUDIO_U8_OFF_VOLUME 128-16
#define AUDIO_FREQUENCY 44100
#define AUDIO_FORMAT AUDIO_U8
#define AUDIO_CHANNELS 1
#define AUDIO_BUFTYP uint8_t
#define AUDIO_BUFSIZE_KB 500

// A much shorter value was found to be necessary required when tested on
// Windows 2000, audio is distorted if the value is too high. The short
// value also seems to work fine under Linux but have used seperate values
// here so that the value can be tweaked to suit the OS.  This may vary
// depending on other OS used and possibly the sound card hardware used.
//
// The number used here must be a power of 2.
//
// MINGW (W2000) - 64 is not enough for when constant tones are generated.
//               - 256 distortion starts.
//               - 128 is about the best compromise.
//
#ifdef MINGW
#define AUDIO_SAMPLES 128
#else
#define AUDIO_SAMPLES 512
#endif

SDL_AudioSpec wanted;
extern void fill_audio(void *udata, Uint8 *stream, int len);

// the sound buffer time capacity in seconds is equal to the size of the
// buffer divided by the AUDIO_FREQUENCY value. A 500K buffer, and an
// AUDIO_FREQUENCY value of 44100 is 11.60 seconds (t=512000/44100)
static AUDIO_BUFTYP sndbuff[1024 * AUDIO_BUFSIZE_KB];

int audio_silence=1;
int dobuffering;

static int audio_count;
static int audio_outpos;
static int audio_inpos;

static UINT8 speaker;

static uint64_t cycles_now;
static uint64_t cycles_before;
static uint64_t cycles_elapsed;

static UINT8 sound_before;

extern uint64_t z80_cycles;
extern clock_t ticksound;

//==============================================================================
// Sound Initialise
//
//   pass: void
// return: int			0 if success, -1 if error
//==============================================================================
int sound_init (void)
{
// set the audio format desired
 wanted.freq = AUDIO_FREQUENCY;
 wanted.format = AUDIO_FORMAT;
 wanted.channels = AUDIO_CHANNELS;
 wanted.samples = AUDIO_SAMPLES;
 wanted.callback = fill_audio;
 wanted.userdata = NULL;

// open the audio device, forcing the desired format
 if (SDL_OpenAudio(&wanted, NULL) < 0)
    {
     fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
     return(-1);
    }

 return 0;
}

//==============================================================================
// The audio function callback function declared in for wanted.callback
//
//   pass: void *udata
//         Uint8 *stream	A pointer to the audio buffer to be filled
//         int len		The length (in bytes) of the audio buffer
// return: void
//==============================================================================
void fill_audio (void *udata, Uint8 *stream, int len)
{
 int count;
 int audio_len1, audio_len2;
 int audio_outnext;

// printf("audio_outpos=%d   audio_count=%d   len=%d\n", audio_outpos, audio_count, len);        

 if (audio_count)
    {
     if (len > audio_count)
        count = audio_count;
     else
        count = len;
        
     if ((count + audio_outpos) <= sizeof(sndbuff))
        {
         audio_len1 = count;
         audio_len2 = 0;
         audio_outnext = audio_outpos + audio_len1;
         if (audio_outnext >= sizeof(sndbuff))
            audio_outnext = 0;
        }
     else   
        {
         audio_len1 = sizeof(sndbuff) - audio_outpos; 
         audio_len2 = count - audio_len1;
         audio_outnext = audio_len2;
        }

// printf("audio_outpos=%d   audio_len1=%d  audio_len2=%d\n", audio_outpos, audio_len1, audio_len2);        

// mix in the audio data, we must do this for audio_len1 and then audio_len2
// if the buffer has wrapped around.
     if (audio_len1)
        {
         SDL_MixAudio(stream, sndbuff + audio_outpos, audio_len1, SDL_MIX_MAXVOLUME);
         if (audio_len2)
             SDL_MixAudio(stream + audio_len1, sndbuff, audio_len2, SDL_MIX_MAXVOLUME);
         audio_count -= count;
         audio_outpos = audio_outnext;
        }

// printf("len=%d   audio_count=%d\n", len, audio_count);
    }
 else
    if (! audio_silence)       
       {
        audio_silence = 1;    
        SDL_PauseAudio(audio_silence);    // silence the audio
       }
}

//==============================================================================
// Sound de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int sound_deinit (void)
{
 SDL_CloseAudio();

 return 0;
}

//==============================================================================
// Sound reset
//
//   pass: void
// return: int				sound init result
//==============================================================================
int sound_reset (void)
{
// lock the audio to stop the call back thread happening while we operate on
// shared data.

 SDL_CloseAudio();

 SDL_LockAudio();
 
 audio_count = 0;
 audio_outpos = 0;
 audio_inpos = 0;

 speaker = 0;
 cycles_now = 0;
 cycles_before = 0;
 cycles_elapsed = 0;
 sound_before = 0;

 ticksound = 0;
 
 audio_silence = 1;
 dobuffering = 0;
 SDL_PauseAudio(audio_silence);    // silence the audio

// un-lock the audio to allow the call back thread to occur.
 SDL_UnlockAudio();

 return sound_init();
}

//==============================================================================
// Sound write
//
//   pass: UINT8 data
// return: void
//==============================================================================
void sound_w (UINT8 data)
{
 int count;
 int audio_len1;
 int audio_len2;
 UINT8 sound_now;

// mask of the speaker bit
 sound_now = data & 0x40;

// Only do something if the speaker state changes.
 if (sound_now != sound_before)
    {
     sound_before = sound_now;

// calculate the elapsed time since the last change of the speaker
// output,  this will be used to determine the amount of data to be
// placed into the buffer.
     cycles_now = z80_cycles + mz80GetElapsedTicks(0); // use 0 !
     cycles_elapsed = cycles_now - cycles_before;
     cycles_before = cycles_now;

// Use a constant value here, determines the number of repeat values to be
// placed into the buffer at the desired AUDIO_FREQUENCY chosen. It is based
// on 1 channel of 8 bit sound samples.
#ifdef XTAL3375
     count = (cycles_elapsed / 74);
#else
     count = (cycles_elapsed / 44);     
#endif

// This value determines when a change in speaker state should be ignored.
// The value should be kept high enough for low frequency sound but low
// enough to allow the audio to re-sync to the Z80 CPU speed.  The time
// is t=value/44100 (44100 being the AUDIO_FREQUENCY rate).
//
// For the sound to be truely emulated would require long time periods of no
// speaker change to also be put out, but this would mean re-syncing will not
// work as intended and a lot of data would also need to be filled in.  Quiet
// periods longer than (count) will therefore rely on the Z80 emulation speed.

     if (count > 4000)			// 4000=11.0 Hz
         return;

// lock the audio to stop the call back thread happening while we operate on
// shared data.
     SDL_LockAudio();

// truncate the number of fill values if not enough room to fit in the buffer
     if ((audio_count + count) > sizeof(sndbuff))
        count = sizeof(sndbuff) - audio_count;

// calculate the buffer fill quantity,  as the buffer is circular in nature
// values are calulated to fill from the current input position and from the
// start of the buffer when wrapping occurs.
     if ((count + audio_inpos) > sizeof(sndbuff))
        {
         audio_len1 = sizeof(sndbuff) - audio_inpos;
         audio_len2 = count - audio_len1;
        }
     else
        {
         audio_len1 = count;
         audio_len2 = 0;
        } 

// get the actual fill value to placed into the buffer.  The value used is
// the value speaker value on the previous occasion but it probably makes
// no differnce anyway.
     if (sound_before)
        speaker = AUDIO_U8_ON_VOLUME;
     else
        speaker = AUDIO_U8_OFF_VOLUME;

// fill the buffer with the speaker value, audio_len1 is the fill length
// for the current audio_inpos and audio_len2 is the fill length from the
// start of the buffer when wrapping occurs.
     if (audio_len1)
        {
         memset(sndbuff + audio_inpos, speaker, audio_len1);
         audio_inpos += audio_len1;
         if (audio_len2)
            {
             memset(sndbuff, speaker, audio_len2);
             audio_inpos = audio_len2;
            } 
        }

// ensure that the audio input position is in kept inbounds
     if ((audio_inpos) >= sizeof(sndbuff))
        audio_inpos = 0;

// adjust the count of samples now in the buffer ready to be played
     audio_count += count;

// every time a re-syncing period occurs (determined by audio_silence = true)
// a small delay occurs before sound starts to play.  This is required when
// running at a real Microbee CPU speed as the sound out is trying to play at
// the real rate too but race conditions means that audio will sometimes not
// be ready on the next call back and so the sound will be become jerky and
// distorted.
     if (audio_silence && ! dobuffering)         
        {
         dobuffering = 1;
         ticksound = -1;             
        }

// un-lock the audio to allow the call back thread to occur.
     SDL_UnlockAudio();            
    }
}

//==============================================================================
// Sound unsilence - start the audio playing after a buffering period has
// expired.
//
//   pass: void
// return: void
//==============================================================================
void sound_unsilence (void)
{
 dobuffering = 0;
 audio_silence = 0;    
 SDL_PauseAudio(audio_silence);    // un-silence audio
}
#endif

==> ubee512-1.3.0.tar.gz: ubee512-1.3.0/src/sound.c <==
//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                               Sound module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// This module is used to emulate the internal speaker using SDL.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.1.0 - 13 August 2007, S.J.Kay
// - Changes to the volume level to allowing level to be set by user.
// - Added sound_clock function that determines the sound rate playback rate
//   depending on the XTAL frequency and the method chosen.
//
// v1.0.0 - 14 July 2007, S.J.Kay
// - Created a new file to implement the sound emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#ifdef use_audio

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "SDL.h"
#include "sound.h"
#include "ubee512.h"
#include "mz80.h"

#define AUDIO_U8_VOLUME 16
#define AUDIO_FREQUENCY 44100
#define AUDIO_FORMAT AUDIO_U8
#define AUDIO_CHANNELS 1
#define AUDIO_BUFTYP uint8_t
#define AUDIO_BUFSIZE_KB 500

// A much shorter value was found to be necessary required when tested on
// Windows 2000, audio is distorted if the value is too high. The short
// value also seems to work fine under Linux but have used seperate values
// here so that the value can be tweaked to suit the OS.  This may vary
// depending on other OS used and possibly the sound card hardware used.
//
// The number used here must be a power of 2.
//
// MINGW (W2000) - 64 is not enough for when constant tones are generated.
//               - 256 distortion starts.
//               - 128 is about the best compromise.
//
#ifdef MINGW
#define AUDIO_SAMPLES 128
#else
#define AUDIO_SAMPLES 512
#endif

SDL_AudioSpec wanted;
extern void fill_audio(void *udata, Uint8 *stream, int len);

// the sound buffer time capacity in seconds is equal to the size of the
// buffer divided by the AUDIO_FREQUENCY value. A 500K buffer, and an
// AUDIO_FREQUENCY value of 44100 is 11.60 seconds (t=512000/44100)
static AUDIO_BUFTYP sndbuff[1024 * AUDIO_BUFSIZE_KB];

int dobuffering;

int audio_silence=1;
int audio_volume = AUDIO_U8_VOLUME;
static int audio_count;
static int audio_outpos;
static int audio_inpos;
static int audio_divval;
static int audio_off_cmp;

static UINT8 speaker;

static uint64_t cycles_now;
static uint64_t cycles_before;
static uint64_t cycles_elapsed;

static UINT8 sound_before;

extern int sound;

extern uint64_t z80_cycles;
extern clock_t ticksound;

//==============================================================================
// Sound Initialise
//
//   pass: void
// return: int			0 if success, -1 if error
//==============================================================================
int sound_init (void)
{
// set the audio format desired
 wanted.freq = AUDIO_FREQUENCY;
 wanted.format = AUDIO_FORMAT;
 wanted.channels = AUDIO_CHANNELS;
 wanted.samples = AUDIO_SAMPLES;
 wanted.callback = fill_audio;
 wanted.userdata = NULL;

// open the audio device, forcing the desired format
 if (SDL_OpenAudio(&wanted, NULL) < 0)
    {
     fprintf(stderr, "Couldn't open audio: %s\n", SDL_GetError());
     return(-1);
    }

 return 0;
}

//==============================================================================
// The audio function callback function declared in for wanted.callback
//
//   pass: void *udata
//         Uint8 *stream	A pointer to the audio buffer to be filled
//         int len		The length (in bytes) of the audio buffer
// return: void
//==============================================================================
void fill_audio (void *udata, Uint8 *stream, int len)
{
 int count;
 int audio_len1, audio_len2;
 int audio_outnext;

// printf("audio_outpos=%d   audio_count=%d   len=%d\n", audio_outpos, audio_count, len);        

 if (audio_count)
    {
     if (len > audio_count)
        count = audio_count;
     else
        count = len;
        
     if ((count + audio_outpos) <= sizeof(sndbuff))
        {
         audio_len1 = count;
         audio_len2 = 0;
         audio_outnext = audio_outpos + audio_len1;
         if (audio_outnext >= sizeof(sndbuff))
            audio_outnext = 0;
        }
     else   
        {
         audio_len1 = sizeof(sndbuff) - audio_outpos; 
         audio_len2 = count - audio_len1;
         audio_outnext = audio_len2;
        }

// printf("audio_outpos=%d   audio_len1=%d  audio_len2=%d\n", audio_outpos, audio_len1, audio_len2);        

// mix in the audio data, we must do this for audio_len1 and then audio_len2
// if the buffer has wrapped around.
     if (audio_len1)
        {
         SDL_MixAudio(stream, sndbuff + audio_outpos, audio_len1, SDL_MIX_MAXVOLUME);
         if (audio_len2)
             SDL_MixAudio(stream + audio_len1, sndbuff, audio_len2, SDL_MIX_MAXVOLUME);
         audio_count -= count;
         audio_outpos = audio_outnext;
        }

// printf("len=%d   audio_count=%d\n", len, audio_count);
    }
 else
    if (! audio_silence)       
       {
        audio_silence = 1;    
        SDL_PauseAudio(audio_silence);    // silence the audio
       }
}

//==============================================================================
// Sound de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int sound_deinit (void)
{
 SDL_CloseAudio();

 return 0;
}

//==============================================================================
// Sound reset
//
//   pass: void
// return: int				sound init result
//==============================================================================
int sound_reset (void)
{
// lock the audio to stop the call back thread happening while we operate on
// shared data.

 SDL_CloseAudio();

 SDL_LockAudio();
 
 audio_count = 0;
 audio_outpos = 0;
 audio_inpos = 0;

 speaker = 0;
 cycles_now = 0;
 cycles_before = 0;
 cycles_elapsed = 0;
 sound_before = 0;

 ticksound = 0;
 
 audio_silence = 1;
 dobuffering = 0;
 SDL_PauseAudio(audio_silence);    // silence the audio

// un-lock the audio to allow the call back thread to occur.
 SDL_UnlockAudio();

 return sound_init();
}

//==============================================================================
// Sound write
//
//   pass: UINT8 data
// return: void
//==============================================================================
void sound_w (UINT8 data)
{
 int count;
 int audio_len1;
 int audio_len2;
 UINT8 sound_now;

 if (! sound)		// if sound is disabled
    return;

// mask of the speaker bit
 sound_now = data & 0x40;

// Only do something if the speaker state changes.
 if (sound_now != sound_before)
    {
     sound_before = sound_now;

// calculate the elapsed time since the last change of the speaker
// output,  this will be used to determine the amount of data to be
// placed into the buffer.
     cycles_now = z80_cycles + mz80GetElapsedTicks(0); // use 0 !
     cycles_elapsed = cycles_now - cycles_before;
     cycles_before = cycles_now;

// Use a constant value here, determines the number of repeat values to be
// placed into the buffer at the desired AUDIO_FREQUENCY chosen. It is based
// on 1 channel of 8 bit sound samples.
     count = cycles_elapsed / audio_divval;

// This value determines when a change in speaker state should be ignored.
// The value should be kept high enough for low frequency sound but low
// enough to allow the audio to re-sync to the Z80 CPU speed.  The time
// is t=value/44100 (44100 being the AUDIO_FREQUENCY rate).
//
// For the sound to be truely emulated would require long time periods of no
// speaker change to also be put out, but this would mean re-syncing will not
// work as intended and a lot of data would also need to be filled in.  Quiet
// periods longer than (count) will therefore rely on the Z80 emulation speed.

     if (count > audio_off_cmp)			// 4000=11.0 Hz @ 3.375 MHz
         return;

// lock the audio to stop the call back thread happening while we operate on
// shared data.
     SDL_LockAudio();

// truncate the number of fill values if not enough room to fit in the buffer
     if ((audio_count + count) > sizeof(sndbuff))
        count = sizeof(sndbuff) - audio_count;

// calculate the buffer fill quantity,  as the buffer is circular in nature
// values are calulated to fill from the current input position and from the
// start of the buffer when wrapping occurs.
     if ((count + audio_inpos) > sizeof(sndbuff))
        {
         audio_len1 = sizeof(sndbuff) - audio_inpos;
         audio_len2 = count - audio_len1;
        }
     else
        {
         audio_len1 = count;
         audio_len2 = 0;
        } 

// get the actual fill value to be placed into the buffer.  The value used is
// the speaker value on the previous occasion but it probably makes
// no difference anyway.
     if (sound_before)
        speaker = 128 + audio_volume;
     else
        speaker = 128 - audio_volume;

// fill the buffer with the speaker value, audio_len1 is the fill length
// for the current audio_inpos and audio_len2 is the fill length from the
// start of the buffer when wrapping occurs.
     if (audio_len1)
        {
         memset(sndbuff + audio_inpos, speaker, audio_len1);
         audio_inpos += audio_len1;
         if (audio_len2)
            {
             memset(sndbuff, speaker, audio_len2);
             audio_inpos = audio_len2;
            } 
        }

// ensure that the audio input position is in kept inbounds
     if ((audio_inpos) >= sizeof(sndbuff))
        audio_inpos = 0;

// adjust the count of samples now in the buffer ready to be played
     audio_count += count;

// every time a re-syncing period occurs (determined by audio_silence = true)
// a small delay occurs before sound starts to play.  This is required when
// running at a real Microbee CPU speed as the sound out is trying to play at
// the real rate too but race conditions means that audio will sometimes not
// be ready on the next call back and so the sound will be become jerky and
// distorted.
     if (audio_silence && ! dobuffering)
        {
         dobuffering = 1;
         ticksound = -1;
        }    
            
// un-lock the audio to allow the call back thread to occur.
     SDL_UnlockAudio();            
    }
}

//==============================================================================
// Sound unsilence - start the audio playing after a buffering period has
// expired.
//
//   pass: void
// return: void
//==============================================================================
void sound_unsilence (void)
{
 dobuffering = 0;
 audio_silence = 0;    
 SDL_PauseAudio(audio_silence);    // un-silence audio
}

//==============================================================================
// Sound clock values - set values determined by the XTAL rate,  Microbee
// application software was designed for 2 and 3.75 MHz XTALS so if these are
// required then method should be 0. (proportional)
//
// If the method is 1 then playback will always be at the rate expected for a
// 3.375 MHz XTAL and software regardless of the XTAL frequency.
//
//   pass: int xtal		XTAL frequency in Hz.
//         int method		0=proportional sound rate determined by XTAL
//                              1=sound rate always as if 3.375 MHz XTAL used
// return: void
//==============================================================================
void sound_clock (int xtal, int method)
{
 if (method == SOUND_PROP)
    {
     audio_divval = (int)(((float)xtal / 3375000.0) * 74.0);
     audio_off_cmp = (int)(((float)xtal / 3375000.0) * 4000.0);
    } 
 else
    {
     audio_divval = 74;		// sound is normal regardless of XTAL freq
     audio_off_cmp = 4000;
    } 
}

#endif
